import os

def get_filename(filetype):
  '''
  forces user to input filename if one not detected with filetype == <filetype>

  <filetype> example: '.json'
  '''

  def outer(func):

    def inner(*args, **kwargs):
      print(kwargs)
      print(args)

      filename = kwargs.get('filename')
      if not filename:
        while True:
          filename = input(f'Please enter a valid filename (filetype== {filetype}: ')
          if get_file_parts(filename)[1] == filetype:
            kwargs['filename'] = filename
      
      return func(*args, **kwargs)
    
    return inner
  
  return outer


      



















#region HELPERS FOR DECORATORS

def get_file_parts(file_):
  '''
  file needs appropriate path headed

  returns filename, file_extension
  '''
  print(os.path.splitext(file_))
  return os.path.splitext(file_)



#endregion HELPERS FOR DECORATORS


import cProfile
from functools import wraps
import datetime


def prnt(val, label='undefined label'):
    '''
    Print wrapper for clear logging
    '''
    print(f"\n{label}: {val}\n")


def prnt_label_val(val, label='undefined label'):
    # print wrapper that prints label and then value on next line with indentation
    print(f"{label}\n    {val}")


def print_keys(dict_, depth=0):
    '''
    prints keys recursively at a certain <depth> and indents based on <depth> (times 2)
    '''
    for key in dict_:
        print(f"{' ' * depth * 2}{key}")
        if isinstance(dict_[key], dict):
            print_keys(dict_[key], depth+1)


def print_items(dict_, depth=0):
    for key in dict_:
        print(f"{' ' * depth * 2}{key}")
        if isinstance(dict_[key], dict):
            print_items(dict_[key], depth+1)
        elif isinstance(dict_[key], list):
            print("\n")
            for item_ in dict_[key]:
                print(f"{' ' * (depth + 1) * 2}{item_}")
            print("\n")
        else:
            print(f"\n{' ' * (depth + 1) * 2}{dict_[key]}\n")

def prnt_attrs(obj, *attrs, incl_dir=True):
    if incl_dir:
        prnt(dir(obj))
    for attr in attrs:
        try:
            prnt(getattr(obj, attr), attr)
        except AttributeError:
            prnt(f'{obj} has no attribute: {attr}')

def print_object_attrs(obj, len_limit=1000, try_call_functions=True):

    def print_truncated_label_value(label, value):
        try:
            if len(value) > len_limit:
                prnt_label_val(f'{value[:len_limit]}\n\n...\n\t\tValue truncated to first {len_limit} characters', label)
            else:
                prnt_label_val(value, label)
        except:
            prnt_label_val('No value could be determined', label)


    if try_call_functions:
        for attr in dir(obj):
            if callable(getattr(obj, attr)):
                try:
                    value = getattr(obj, attr)()
                    print_truncated_label_value(attr, value)
                except:
                    prnt_label_val(f'ERROR calling {getattr(obj, attr)}', attr)
            else:
                value = getattr(obj, attr)
                print_truncated_label_value(attr, value)

    else:
        for attr in dir(obj):
            value = getattr(obj, attr)
            print_truncated_label_value(attr, value)


def get_item_with_largest_val(items, attr, attr_type='int'):
    '''
    probably needs rewriting

    input is expected to be an object of objects or a list of objects

    returns object from <item>'s objects that has largest val of <attr>
    '''

    output_item = None
    if isinstance(items, list):
        for obj in items:
            if not output_item:
                output_item = obj

            if attr_type == 'int':
                if int(output_item[attr]) < int(obj[attr]):
                    output_item = obj

    elif isinstance(items, object):

        for key in items:
            if not output_item:
                output_item = items[key]

            if attr_type == 'int':
                if int(output_item[attr]) < int(items[key][attr]):
                    output_item = items[key]

    return output_item


def get_objects_with_attr_val(objects, attr, val):
    '''
    returns list of objects from input list/object <objects> that has an <attr> value of <val>
    '''

    output = []

    if isinstance(objects, object):
        for obj in objects:
            if objects[obj][attr] == val:
                output.append(obj)

    elif isinstance(objects, list):
        for obj in objects:
            if obj[attr] == val:
                output.append(obj)

    return output


def combine_lists(list1, list2):
    # lists must be of same size or list2 must be 1 less than list1
    output = []
    for i in range(len(list1)):
        output.append(list1[i])
        try:
            output.append(list2[i])
        except IndexError:
            pass
    return output

def print_command_history():
    import readline
    for i in range(readline.get_current_history_length()):
        print (readline.get_history_item(i + 1))

def convert_to_iso_8601(datetime_obj):
    # returns iso8601 string
    return datetime_obj.strftime('%Y-%m-%dT%H:%M:%SZ')

def convert_from_iso_8601(iso8601_string):
    # returns datetime obj
    return datetime.datetime.strptime(iso8601_string, '%Y-%m-%dT%H:%M:%SZ')



from ..helpers import *
import cProfile




def cProfile_dec(sort_by='tottime'):

  def outer(func):

    def inner(*args, **kwargs):

      profile = cProfile.Profile()
      profile.enable()

      output = func(*args, **kwargs)

      profile.print_stats(sort=sort_by)
      # sorting methods here:
      # https://docs.python.org/3/library/profile.html#pstats.Stats.sort_stats

      profile.disable()

      return output

    return inner

  return outer


def timeit(func):
  #decorator that wraps function with simple time log

  def inner(*args, **kwargs):

    t0 = time.time()
    prnt(args, 'args')
    prnt(kwargs, 'kwargs')

    output = func(*args, **kwargs)

    prnt(time.time() - t0, f"Time taken for {func.__name__}")

    return output

  return inner


@cProfile_dec()
def prof_test():
  j=1
  for i in range(10000000):
    j*=1.00000001

  return j


import os
import cProfile
import time
from ..helpers import *
import logging
from ..mongo_engine import *
from datetime import datetime


def exponential_backoff(
        start_point=1, rate_gain=2, rate_type='exponential', max_time=100):
    '''
    for respecting other people's wishes

    the function that is passed in should raise an error (explicitly or implicitly) when output is not as desired
    for example if doing a url request and you get forbidden status (403), raise an error upon reception of that request code and
    this wrapper will handle exponentially backing off
    '''
    start_point = start_point
    rate_gain = rate_gain
    rate_type = rate_type
    max_time = max_time
    logging.basicConfig(
        filename='/var/log/usr/exponential-backoffs.log', level=logging.INFO,
        format='%(asctime)s %(message)s')
    error = None

    def outer(func):

        def inner(*args, **kwargs):
            sleep = 0
            while True:
                try:
                    if sleep > max_time:
                        return logging.info(
                            f"Function {func.__name__} with args {args} and"
                            f" kwargs {kwargs} failed with error:\n       {error}"
                        )
                    output = func(*args, **kwargs)
                    break
                except Exception as e:
                    if rate_type == 'exponential':
                        sleep = sleep * rate_gain or start_point
                    error = e
                    print(error)
                    print(
                        f"Exponentially backing off: waiting {sleep} seconds")
                    time.sleep(sleep)

            return output

        return inner
    return outer


# region *start* EXPONENTIAL BACKOFF PROOF/TESTING

i = 0
dict_ = {4: 'hey'}


@exponential_backoff(rate_gain=4, max_time=4)
def myfunc(a, b, c='this'):
    global i, dict_
    print(i)
    i += 1
    print(dict_[i])
    return i

# endregion *end* EXPONENTIAL BACKOFF PROOF/TESTING


def update_db(interval=24):
    '''
    interval param indicates how long we wait (in hours) before grabbing value from the source
    '''

    interval = interval * 3600

    def outer(func):

        def inner(*args, **kwargs):

            # if datetime.datetime.now() - db_entry.last_updated

            return func(*args, **kwargs)

        return inner

    return outer


def memo(func):

    cache = {}

    def inner(*args):
        try:
            return cache[args]
        except KeyError:
            cache[args] = result = func(*args)
            return result
        except TypeError:
            return func(*args)

    return inner


def db_lookup(db, coll, key, last_updated=24):
    '''
    This is essentially a memoize function for database lookups that sets the value if it doesn't \
      already exist or has surpassed the allotted last_updated time

    IMPORTANT: In order to use this decorator, one must specify a selection_filter kwarg \
      in the function that is wrapped by this decorator. The selection_filter value
      is a standard mongo filter selection object.
      This is necessary because there is no way to consistently pass dynamic values
      prior to the decorator being called

      So we find the document based on this and then update the desired <key> to the
      return value of our function

    EXAMPLE:

      @db_lookup('scraped_data', 'full_html', 'html')
      function get_html(self, selection_filter={'url': 'example_url'}):
        # return the html based on a get from <selection_filter[url]>

      That example function will first check the database for where url matches the example_url \
        and then if it doesnt exist or if its been greater than 24 hours since the field was last updated
        then it will return the html and db_lookup will automatically handle changing the last_updated field
        and the returned html to the html field

    IF THE VALUE DOES NOT EXIST IN THE DB THEN WE AUTOMATICALLY INSERT ONE; \
      ONLY THE VALUES PRESENT IN THE KEY AND THE SELECTION FILTER WILL
      BE DEFINED IN THE NEWLY CREATED DOCUMENT; OH AND last_updated will be included


    Parameters
    ----------

    db: str
      The MongoDB database you wish to connect to

    coll: str
      The MongoDB database collection you wish to check data from

    key: str
      The key value you wish to return

    last_updated: int or float
      How long (in hours) before a cached value is re-updated
    '''
    if last_updated:
        last_updated_seconds = last_updated * 3600
    db_ = MongoEngine(db=db, collection=coll)

    def outer(func):

        def inner(*args, **kwargs):
            document = db_.collection.find_one(kwargs['selection_filter'])
            if document:
                if last_updated_seconds:
                    now_ts = datetime.utcnow().timestamp()
                    document_ts = document['last_updated'].timestamp()
                    if now_ts - document_ts > last_updated_seconds:

                        value = func(*args, **kwargs)
                        db_.update_value(
                            kwargs['selection_filter'],
                            key, value, 'one')

                        return value

                print('Returning cached version')
                return document[key]

            else:
                value = func(*args, **kwargs)

                # instantiate new entry as the selection args
                new_entry = kwargs['selection_filter']

                # update that entry with our newly generated value and current time as last_updated
                new_entry.update({
                    'last_updated': datetime.utcnow(),
                    key: value
                })

                # add to the db
                db_.add_entry(new_entry)

                # return the value we initially were looking for
                return value

        return inner

    return outer
